#!/usr/bin/env python
"""
The Pyphysim library provides many useful classes and functions to
perform Monte Carlo simulations of communications systems.
"""

# The short X.Y version.
__version__ = 0.4
