#!/usr/bin/env python
"""Package with MATLAB related functions.

Examples: conversions for the print output from python to MATLAB and
vice-verse. That way you can print a numpy array, copy the output and past
in MATLAB for further analysis.

"""

from . import python2MATLAB
